#!/usr/bin/python
# -*- coding: iso-8859-1 -*-
# $Id: bloxxor.py,v 1.11 2007-10-12 02:54:07 vah Exp $

"""
Copyright (c) 2007  Vin Helminen
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:
   1. Redistributions of source code must retain the above
      copyright notice, this list of conditions and the following
      disclaimer.
   2. Redistributions in binary form must reproduce the above copyright
      notice, this list of conditions and the following disclaimer in
      the documentation and/or other materials provided with the
      distribution.

THIS SOFTWARE IS PROVIDED BY VIN HELMINEN ``AS IS'' AND ANY EXPRESS
OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED. IN NO EVENT SHALL VIN HELMINEN OR CONTRIBUTORS BE
LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
THE POSSIBILITY OF SUCH DAMAGE.
"""

import os, sys, random, colorsys
import pygame

KEY_QUIT = pygame.K_ESCAPE
KEY_FULLSCREEN = pygame.K_f
KEY_TOGGLE_SOUND = pygame.K_s

KEY_LEFT = pygame.K_LEFT
KEY_RIGHT = pygame.K_RIGHT
KEY_SPEEDUP = pygame.K_UP
KEY_FIRE = pygame.K_LCTRL

KEY_SCREENSHOT = pygame.K_F12

MAP_WIDTH = 16
MAP_HEIGHT = 20
TILE_SIZE = 24
SCREEN_WIDTH = MAP_WIDTH*TILE_SIZE
SCREEN_HEIGHT = MAP_HEIGHT*TILE_SIZE
WINDOW_WIDTH = 640
WINDOW_HEIGHT = 480
SCREEN_X_OFFSET = (WINDOW_WIDTH-SCREEN_WIDTH)//2
SCREEN_Y_OFFSET = (WINDOW_HEIGHT-SCREEN_HEIGHT)//2
FULLSCREEN = False
FRAMERATE = 30
ENABLE_SCREENSHOTS = False

BLOXXOR = (
    ((1, 1, 1, 1),),
    
    ((1, 1),
     (1, 0),
     (1, 0),
     (1, 0),
     (1, 0)),
    
    ((1, 1),
     (0, 1),
     (0, 1),
     (0, 1),
     (0, 1)),
    
    ((1, 1, 1, 1, 1),
     (0, 0, 0, 0, 1)),
    
    ((1, 1, 1, 1, 1),
     (1, 0, 0, 0, 0)),
    
    ((1, 1, 1),
     (0, 1, 0),
     (0, 1, 0)),
    
    ((1, 1, 1),
     (1, 0, 0),
     (1, 0, 0)),
    
    ((1, 1, 1),
     (0, 0, 1),
     (0, 0, 1)),
    
    ((1, 1, 1, 1),
     (1, 0, 0, 1),
     (1, 0, 0, 1)),

    ((1, 1, 1, 1, 1),
     (0, 0, 1, 0, 0)),
    )

def HSV2RGB(h, s, v):
    r, g, b = colorsys.hsv_to_rgb((1./360)*h, 0.01*s, 0.01*v)
    return int(255*r), int(255*g), int(255*b)

COLORS = [(HSV2RGB(i, 90, 90), HSV2RGB(i, 20, 100), HSV2RGB(i, 60, 30)) for i in range(0, 360,360//len(BLOXXOR))]

COLORKEY = (255,0,255)

# states
INTRO, PLAY, WARP, GAMEOVER = range(4)

class Stars(pygame.Surface):
    def __init__(self):
        self.height = 2*SCREEN_HEIGHT
        pygame.Surface.__init__(self, (SCREEN_WIDTH, self.height), 0, screen)
        self.offset = 0

        for y in range(0, self.height):
            x = random.randint(-SCREEN_WIDTH, SCREEN_WIDTH)
            if x >= 0:
                c = random.randint(65,200)
                self.set_at((x, y), (c,c,c))

    def scroll(self, pixels):
        self.offset = (self.offset - pixels) % self.height

    def blitTo(self, rect=(0,0,SCREEN_WIDTH,SCREEN_HEIGHT)):
        x, y, w, h = rect
        offset = y + int(self.offset)
        h1 = self.height - offset
        if h1 < 0:
            h1 = 0
            h2 = h
        else:
            h1 = min(h, h1)
            h2 = h - h1
            screen.blit(self, (x, y), (x, offset, w, h1))
        if h2:
            screen.blit(self, (x, y + h1), (x, offset + h1 - self.height, w, h2))        

class Scroller(pygame.Surface):
    def __init__(self):
        pygame.Surface.__init__(self, (SCREEN_WIDTH, SCREEN_HEIGHT+TILE_SIZE), 0, screen)
        self.set_colorkey(COLORKEY)
        self.reset()
        
    def reset(self):
        self.offset = TILE_SIZE
        self.fill(COLORKEY)

    def clear(self, rect):
        self.fill(COLORKEY, rect)

    def scroll(self, pixels):
        self.offset -= pixels
        if self.offset >= 0:
            return 0

        h = TILE_SIZE*( int(TILE_SIZE-self.offset) // TILE_SIZE )
        self.offset += h
        self.set_colorkey(None)
        self.blit(self, (0, h), (0, 0, SCREEN_WIDTH, SCREEN_HEIGHT))
        self.set_colorkey(COLORKEY)
        self.fill(COLORKEY, (0, 0, SCREEN_WIDTH, h))
        return h

    def blitTo(self, rect=(0,0,SCREEN_WIDTH,SCREEN_HEIGHT)):
        x, y, w, h = rect
        screen.blit(self, (x, y), (x, y+int(self.offset), w, h))

class Blob(pygame.Surface):
    def __init__(self, source, rect):
        x, y, w, h = rect
        pygame.Surface.__init__(self, (w, h), 0, screen)
        self.blit(source, (0,0), rect)
        self.alpha = 255
        self.position = x, y - scroller.offset

        blobs.append(self)

    def blitTo(self, speed):
        self.alpha -= speed
        if self.alpha <= 0: return False
        a = int(self.alpha)
        if a & 3 == 0: return True
        self.set_alpha(a)
        screen.blit(self, self.position)
        return True
    
class World(object):
    def __init__(self):
        self.reset()
        self.counter = 0

    def reset(self):
        self.offset = 0
        self.data = [[0]*MAP_WIDTH for y in range(MAP_HEIGHT)]
        self.closest = [-1000]*MAP_WIDTH
        scroller.reset()

    def scroll(self, pixels):
        assert(pixels % TILE_SIZE==0)
        rows = pixels // TILE_SIZE
        if self.offset < rows:
            self.offset += 5
            new = [[0]*MAP_WIDTH for y in range(5)]
            self.data = new + self.data
            if level:
                class BreakOut(Exception):
                    pass
                bloxxor = list(enumerate(BLOXXOR))
                c = 0
                while bloxxor and c < level:
                    c += 1
                    self.counter = (self.counter + 512) & 16777211
                    b = random.randint(0, len(bloxxor)-1)
                    a, blox = bloxxor[b]
                    a += self.counter
                    
                    h = len(blox)
                    w = len(blox[0])
                    
                    def check(ox, oy):
                        for j in range(h):
                            y = oy + j
                            for i in range(w):
                                x = ox + i
                                if blox[j][i] and new[y][x]:
                                    return False
                        return True

                    def put(ox, oy):
                        for j in range(h):
                            y = oy + j
                            for i in range(w):
                                x = ox + i
                                if blox[j][i]:
                                    new[y][x] = a

                    ox = random.randint(0, MAP_WIDTH - w)
                    try:
                        for y in range(5-h):
                            for x in range(ox, MAP_WIDTH - w + 1):
                                if check(x, y):
                                    put(x, y)
                                    raise BreakOut
                            for x in range(0, ox):
                                if check(x, y):
                                    put(x, y)
                                    raise BreakOut
                    except BreakOut:
                        continue

                    del bloxxor[b] # blox does not fit anywhere

        self.offset -= rows
        data = self.data = self.data[:(MAP_HEIGHT+self.offset)]

        for y in range(0, rows):
            j = y * TILE_SIZE
            yoff = y + self.offset
            row = data[yoff]
            for x in range(MAP_WIDTH):
                a = row[x]
                if a:
                    i = x * TILE_SIZE
                    n = self.checkNeighbours(x, yoff, a)
                    self.drawBlox(i, j, a, n)

        self.updateClosest()

    def updateClosest(self):
        data = self.data
        self.closest = []
        for x in range(MAP_WIDTH):
            closest = -1000
            for y in range(len(data)-1, -1, -1):
                if data[y][x]:
                    closest = y - self.offset
                    break
            self.closest.append(closest)


    def checkNeighbours(self, x, y, a):
        n = 0
        if y and self.data[y-1][x] == a:
            n |= 0x1
        if x and self.data[y][x-1] == a:
            n |= 0x2
        if x+1 < MAP_WIDTH and self.data[y][x+1] == a:
            n |= 0x4
        if y+1 < len(self.data) and self.data[y+1][x] == a:
            n |= 0x8
        return n
            
    def drawBlox(self, x, y, a, n):
        color, highcolor, lowcolor = COLORS[a & 511]
        scroller.fill(color, (x, y, TILE_SIZE, TILE_SIZE))
        leftspace = int(not n & 0x2)
        rightspace = int(not n & 0x4)
        topspace = int(not n & 0x1)
        bottomspace = int(not n & 0x8)
        w = TILE_SIZE-leftspace-rightspace
        h = TILE_SIZE-topspace-bottomspace
        s = TILE_SIZE-2
        if topspace:
            scroller.fill(highcolor, (x+leftspace, y+1, w, 1))
        if leftspace:
            scroller.fill(highcolor, (x+1, y+topspace, 1, h))
        if rightspace:
            scroller.fill(lowcolor, (x+s, y+topspace, 1, h))
        if bottomspace:
            scroller.fill(lowcolor, (x+leftspace, y+s, w, 1))

    def hit(self, x, y):
        assert(self.closest[x] == y)
        if y < 0 or y >= MAP_HEIGHT: return
        data = self.data
        y += self.offset
        a = data[y][x]
        data[y+1][x] = a

        def p(x, y):
            if x < 0 or x >= MAP_WIDTH or data[y][x] != a: return
            n = self.checkNeighbours(x, y, a)
            self.drawBlox(x*TILE_SIZE, (y-self.offset)*TILE_SIZE, a, n)

        p(x, y)
        p(x, y+1)
        p(x-1, y+1)
        p(x+1, y+1)
        
        self.closest[x] += 1

        def checkBox(x, y):
            # find bottom left corner
            while True:
                while x and data[y][x-1] == a: x -= 1
                if y and data[y-1][x] == a:
                    y -= 1
                    continue
                break

            # get width and height
            ox = x
            oy = y
            while x < MAP_WIDTH and data[oy][x] == a: x += 1
            while y < len(data) and data[y][ox] == a: y += 1
            w = x - ox
            h = y - oy
            # check for properly filled rectangle
            for y in range(oy, oy+h):
                if ox and data[y][ox-1] == a: return False
                if ox+w < MAP_WIDTH and data[y][ox+w] == a: return False
                for x in range(ox, ox+w):
                    if data[y][x] != a: return False
            y = oy + h
            if y < len(data):
                for x in range(ox, ox+w):
                    if data[y][x] == a: return False
            y = oy - 1
            if y:
                for x in range(ox, ox+w):
                    if data[y][x] == a: return False

            # clear rectangle
            for y in range(oy, oy+h):
                for x in range(ox, ox+w):
                    data[y][x] = 0
    
            oy -= self.offset
            ox *= TILE_SIZE
            oy *= TILE_SIZE
            w *= TILE_SIZE
            h *= TILE_SIZE
            r = (ox, oy, w, h)
            Blob(scroller, r)
            scroller.clear(r)

            return True

        if checkBox(x, y):
            self.updateClosest()
            playSound(sndhit)

    
class Bullet(object):
    def __init__(self):
        self.active = False
        self.colorid = len(COLORS)

    def launch(self, x, y):
        self.x = x
        self.y = y
        self.active = True

    def move(self):
        self.y -= 1
        closest = world.closest[self.x]
        if closest < 0:
            if self.y <= 0:
                self.active = False
        elif self.y <= closest:
            world.hit(self.x, closest)
            self.active = False
        
    def paint(self):
        if self.colorid <= 0: self.colorid = len(COLORS)
        self.colorid -= 1
        x = self.x*TILE_SIZE
        y = self.y*TILE_SIZE - scroller.offset + 1
        screen.fill(COLORS[self.colorid][0], (x, y, TILE_SIZE, TILE_SIZE))
        screen.fill((0,0,0), (x+2, y+2, TILE_SIZE-4, TILE_SIZE-4))

        
class Player(object):
    def __init__(self):
        self.yspeed = 1
        self.bullets = [Bullet() for i in range(4)]
        self.reset()
        self.sprite = loadImage('player.bmp')
        self.sprite.set_colorkey((255,0,255))
        
    def reset(self):
        self.x = MAP_WIDTH // 2
        self.sx = TILE_SIZE*self.x - TILE_SIZE // 2
        self.movecounter = 0
        self.left = False
        self.right = False
        self.xspeed = 0
        self.yspeed = 1
        for b in self.bullets:
            b.active = False
        self.level = 0
        self.percent = 0

    def move(self):
        if self.left:
            if self.right:
                self.xspeed = 0
            else:
                self.xspeed = -1
        elif self.right:
            self.xspeed = 1
        else:
            self.xspeed = 0
            
        if self.xspeed:
            if self.movecounter <= 0:
                self.x += self.xspeed
                if self.x >= MAP_WIDTH: self.x = MAP_WIDTH - 1
                elif self.x < 0: self.x = 0
                self.movecounter = FRAMERATE//15
            else:
                self.movecounter -= 1
        else:
            self.movecounter = 0

        x = TILE_SIZE*self.x - TILE_SIZE // 2
        self.sx = (x + self.sx)//2

        for b in self.bullets:
            if b.active: b.move()

    def paint(self):
        x = self.x
        y = world.closest[x]
        if y < 0:
            y = 0
        else:
            y += 1
            y *= TILE_SIZE
            y -= scroller.offset - 1
        x *= TILE_SIZE
        screen.fill((255,255,255), (x, y, TILE_SIZE, TILE_SIZE))
        screen.fill((0,0,0), (x+1, y+1, TILE_SIZE-2, TILE_SIZE-2))

        for b in self.bullets:
            if b.active: b.paint()
            
        y = SCREEN_HEIGHT-2*TILE_SIZE
        screen.blit(self.sprite, (self.sx, y))

        screen.fill((120,120,200), (0, y-2, SCREEN_WIDTH, 2))

    
    def fire(self):
        for b in self.bullets:
            if not b.active:
                b.launch(self.x, MAP_HEIGHT-2)
                playSound(sndfire)
                break

def centerBlit(dst, src, x=None, y=None):
    dw, dh = dst.get_size()
    sw, sh = src.get_size()
    if x == None: x = (dw - sw)//2
    if y == None: y = (dh - sh)//2
    dst.blit(src, (x, y))


def initWindow(fullscreen):
    global screen, window
    if fullscreen: flags = pygame.FULLSCREEN
    else: flags = 0
    window = pygame.display.set_mode((WINDOW_WIDTH, WINDOW_HEIGHT), flags)
    for y in range(0, WINDOW_HEIGHT):
        c = random.randint(80,100)
        window.fill((c,c,c),(0,y,WINDOW_WIDTH,1))

    pygame.display.flip()

    screen = window.subsurface((SCREEN_X_OFFSET,SCREEN_Y_OFFSET,SCREEN_WIDTH, SCREEN_HEIGHT))
    screen.fill((0,0,0))
    
    def scratch(x0, y0, x1, y1):
        x = random.randint(x0, x1)
        y = random.randint(y0, y1)
        for i in range(32000):
            c = random.randint(20,80)
            window.set_at((x, y), (c,c,c, 53))
            x += random.choice([-1,1])
            y += random.choice([-1,1])
            if x < x0 or x > x1 or y < y0 or y > y1: break

    for i in range(4):
        scratch(0, 0, SCREEN_X_OFFSET-1, WINDOW_HEIGHT-1)
        scratch(WINDOW_WIDTH - SCREEN_X_OFFSET, 0, WINDOW_WIDTH-1, WINDOW_HEIGHT-1)

    font = pygame.font.SysFont('serif', max(12,TILE_SIZE))
    title = font.render('BloXXoR', 1, (0,0,0))
    window.blit(title, ((SCREEN_X_OFFSET - title.get_width())//2+1,3))
    title = font.render('BloXXoR', 1, (0,200,255))
    window.blit(title, ((SCREEN_X_OFFSET - title.get_width())//2,2))

    font = pygame.font.SysFont('sans-serif', 12)
    copyright = font.render('Copyright  2007', 1, (30,30,30))
    window.blit(copyright, ((SCREEN_X_OFFSET - copyright.get_width())//2, WINDOW_HEIGHT-2*copyright.get_height()-2))
    copyright = font.render('Vin Helminen', 1, (30,30,30))
    window.blit(copyright, ((SCREEN_X_OFFSET - copyright.get_width())//2, WINDOW_HEIGHT-copyright.get_height()-2))

    tmp = screen.copy()
    tmp.set_alpha(80)
    window.blit(tmp, (SCREEN_X_OFFSET-2, SCREEN_Y_OFFSET))
    window.blit(tmp, (SCREEN_X_OFFSET-1, SCREEN_Y_OFFSET))
    tmp.fill((160,160,160))
    window.blit(tmp, (SCREEN_X_OFFSET+2, SCREEN_Y_OFFSET))
    window.blit(tmp, (SCREEN_X_OFFSET+1, SCREEN_Y_OFFSET))

    tmp = pygame.Surface((SCREEN_X_OFFSET-8,2*max(12, TILE_SIZE)+4), 0, screen)
    tmp.fill((0,0,0))
    window.blit(tmp, (WINDOW_WIDTH - SCREEN_X_OFFSET + 4, 4))
    tmp.set_alpha(64)
    window.blit(tmp, (WINDOW_WIDTH - SCREEN_X_OFFSET + 3, 3))
    tmp.fill((160,160,160))
    window.blit(tmp, (WINDOW_WIDTH - SCREEN_X_OFFSET + 5, 5))
    
    pygame.display.flip()

def getFilePath(name):
    return os.path.join(os.path.dirname(__file__), 'data', name)

def loadSound(name, volume=None):
    print 'Loading:', name
    sound = pygame.mixer.Sound(getFilePath(name))
    if volume: sound.set_volume(volume)
    return sound

def loadImage(name):
    print 'Loading:', name
    return pygame.image.load(getFilePath(name))


def renderTitle():
    def name(x): return pygame.key.name(x).title()
    left = 'Move Left:', 'Move Right:', 'Fire:', 'Speedup:', 'Toggle Sound:', 'Toggle Fullscreen:', 'Quit:'
    right = (name(KEY_LEFT), name(KEY_RIGHT), name(KEY_FIRE), name(KEY_SPEEDUP), name(KEY_TOGGLE_SOUND), name(KEY_FULLSCREEN), name(KEY_QUIT))

    left = [font.render(t, 1, (0,180,200)) for t in left]
    right = [font.render(t, 1, (200,200,200)) for t in right]

    leftwidth = max([t.get_width() for t in left])
    rightwidth = max([t.get_width() for t in right])
    height = font.get_linesize()

    title = titlefont.render('BloXXoR', 1, (0, 200,255))

    surface = pygame.Surface((max(title.get_width(), leftwidth+rightwidth), title.get_height()*2 + height*len(left)), pygame.SRCALPHA, 32)

    centerBlit(surface, title, y = 0)
    y = 2*title.get_height()
    for l, r in zip(left, right):
        surface.blit(l, (0, y))
        surface.blit(r, (leftwidth + (rightwidth - r.get_width())//2, y))
        y += height
    
    return surface.convert_alpha(screen)
    

def drawScore():
    bgcolor = (0,0,0)
    fgcolor = (200,200,255)
    level = font.render('Level: %d'%player.level, 0, fgcolor)
    percent = font.render('%.1f%%'%player.percent, 0, fgcolor)

    w = SCREEN_X_OFFSET - 8
    h = 2*max(12, TILE_SIZE)+4
    x = WINDOW_WIDTH - SCREEN_X_OFFSET + 4
    y = 4 + (h - level.get_height() - percent.get_height())//2
    window.fill(bgcolor, (x, 4, w, h))
    window.blit(level, (x+2, y))
    window.blit(percent, (WINDOW_WIDTH - 6 - percent.get_width(), y+level.get_height()))


def playSound(sound):
    if soundEnabled:
        sound.play()

print '*** BloXXoR ***'
print 'Copyright (c) 2007 Vin Helminen'
print 'All rights reserved.'
print
print 'Initializing...'
pygame.init()
try:
    if pygame.joystick.get_count() > 0:
        joystick = pygame.joystick.Joystick(0)
        joystick.init()
        print 'Using joystick:', joystick.get_name()
    else:
        print 'Not using joystick'
        joystick = None
    fullscreen = FULLSCREEN
    pygame.display.set_caption('BloXXoR')
    try: pygame.display.set_icon(loadImage('icon_small.png'))
    except: pass
    pygame.mouse.set_visible(False)
    initWindow(fullscreen)
    scroller = Scroller()
    stars = Stars()
    world = World()
    clock = pygame.time.Clock()
    font = pygame.font.SysFont('sans-serif', max(12,TILE_SIZE))
    titlefont = pygame.font.SysFont('serif', max(16,2*TILE_SIZE))
    state = INTRO
    level = 0
    player = Player()
    travel = nextlevel = 1
    title = gametitle = renderTitle()
    gameovertitle = titlefont.render('Game Over', 1, (255,0,0)).convert_alpha(screen)
    warptitle = titlefont.render('Warping in...', 1, (0,200,255)).convert_alpha(screen)
    blobs = []
    soundEnabled = True
    sndstart = loadSound('scifi060.wav')
    sndfire = loadSound('LaserSht.wav', 0.3)
    sndhit = loadSound('ArrwThmp.wav', 0.9)
    sndgameover = loadSound('thunderrumble2.wav')
    sndlevel = loadSound('DeadRing.wav', 0.5)

    if ENABLE_SCREENSHOTS:
        screenshots = 0

    clock.tick(FRAMERATE)

    print 'Ready.'
    while True:
        while True:
            event = pygame.event.poll()
            if not event: break

            if event.type == pygame.QUIT:
                sys.exit(0)

            if event.type == pygame.KEYDOWN:
                if event.key == KEY_QUIT:
                    sys.exit(0)

                elif event.key == KEY_FULLSCREEN:
                    fullscreen = not fullscreen
                    initWindow(fullscreen)

                elif event.key == KEY_TOGGLE_SOUND:
                    soundEnabled = not soundEnabled

                elif ENABLE_SCREENSHOTS and event.key == KEY_SCREENSHOT:
                    screenshots += 1
                    pygame.image.save(window, 'shot%.4d.bmp' % screenshots)
                    
            if state == PLAY:
                if event.type == pygame.KEYDOWN:
                    if event.key == KEY_QUIT:
                        state = INTRO
                        title = renderTitle('BloXXoR')
                    elif event.key == KEY_LEFT:
                        player.left = True
                    elif event.key == KEY_RIGHT:
                        player.right = True
                    elif event.key == KEY_SPEEDUP:
                        player.yspeed += 10
                    elif event.key == KEY_FIRE:
                        player.fire()

                elif event.type == pygame.KEYUP:
                    if event.key == KEY_LEFT:
                        player.left = False
                    elif event.key == KEY_RIGHT:
                        player.right = False
                    elif event.key == KEY_SPEEDUP:
                        player.yspeed -= 10

                elif event.type == pygame.JOYAXISMOTION:
                    if event.axis == 0:
                        player.left = player.right = False
                        if event.value < -0.5:
                            player.left = True
                        elif event.value > 0.5:
                            player.right = True                        
                        

                elif event.type == pygame.JOYBUTTONDOWN:
                    player.fire()

            elif state == INTRO:
                if event.type == pygame.JOYBUTTONDOWN or (event.type == pygame.KEYDOWN and event.key == KEY_FIRE):
                    state = WARP
                    title = warptitle
                    playSound(sndstart)
                    level = 1
                    travel = 0
                    nextlevel = 30*TILE_SIZE
                    random.shuffle(COLORS)
                    player.reset()
                    world.reset()
                    


        if state == WARP or state == GAMEOVER:
            s = 0.5*TILE_SIZE
            stars.scroll(33)
        else:
            s = player.yspeed * TILE_SIZE * 0.0006 * clock.get_time()
            stars.scroll(player.yspeed)
        stars.blitTo()

        for i in range(len(blobs)-1, -1, -1):
            if not blobs[i].blitTo(100*s):
                del blobs[i]

        y = scroller.scroll(s)
        if y: world.scroll(y)
        scroller.blitTo()

        if state == WARP:
            player.paint()

            travel += s
            if travel >= SCREEN_HEIGHT/2.:
                state = PLAY
                travel = 0
                title = None
                
        elif state == PLAY:
            player.move()
            player.paint()

            closest = max(world.closest)
            travel += s
            if travel >= nextlevel or (closest < 0 and travel > 12*TILE_SIZE):
                travel = 0
                playSound(sndlevel)
                level += 1
                nextlevel = 30.*TILE_SIZE*level
                if level > 10:
                    player.yspeed += 0.5
                    
            player.level = level
            player.percent = 100.*travel/nextlevel

            if TILE_SIZE * closest + scroller.offset >= (MAP_HEIGHT - 2)*TILE_SIZE:
                state = GAMEOVER
                title = gameovertitle
                playSound(sndgameover)
                travel = 0

        elif state == GAMEOVER:
            travel += s
            if travel >= SCREEN_HEIGHT:
                title = gametitle
                state = INTRO
                level = 0


        drawScore()

        if title: centerBlit(screen, title)

        #tmp = font.render('%.1f FPS'%clock.get_fps(), 0, (255,0,0))
        #screen.blit(tmp, (1,1))

        pygame.display.flip()    
        clock.tick(FRAMERATE)


except KeyboardInterrupt: pass

print 'Exiting...'
if joystick: joystick.quit()
pygame.quit()
print 'Bye.'
